"""Implementation of pairs for Python."""

from types import ListType as __ListType
from types import TupleType as __TupleType
from symbol import Symbol
import symbol
from error import SchemeError
from operator import add as __add
import unittest


def isNull(x):
    return x == []


def schemeCons(head, rest):
    """Returns the concatentation of head with rest.

    Note: If rest isn't a list, we use a peculiar representation of a
    pair:

        ([head], [rest]).

    The reason for this is to allow set-car! and set-cdr! to work if we decide
    to implement it.
    """
    return ([head], [rest])



def schemeCar(L):
    if type(L) not in (__ListType, __TupleType):
        raise SchemeError, "CAR --- argument is not a pair."
    if len(L) == 0:
        raise SchemeError, "CAR --- cannot take CAR of empty list."
    if type(L) == __ListType: return L[0]
    else: return L[0][0]


def schemeCdr(L):
    if type(L) not in (__ListType, __TupleType):
        raise SchemeError, "CDR --- argument is not a pair."
    if len(L) == 0:
        raise SchemeError, "CDR --- cannot take CDR of empty list."
    if type(L) == __ListType: return L[1:]
    else: return L[1][0]


def schemeCddr(L):
    return schemeCdr(schemeCdr(L))

def schemeCdddr(L):
    return schemeCdr(schemeCdr(schemeCdr(L)))


def schemeCadr(L):
    return schemeCar(schemeCdr(L))

def schemeCaddr(L):
    return schemeCar(schemeCdr(schemeCdr(L)))

def schemeCadddr(L):
    return schemeCar(schemeCdr(schemeCdr(schemeCdr(L))))


def schemeSetCarBang(pair, element):
    if type(pair) != __TupleType:
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[0][0] = element
    return Symbol("ok")


def schemeSetCdrBang(pair, element):
    if type(pair) != __TupleType:
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[1][0] = element
    return Symbol("ok")


def isList(p):
    if not isPair(p): return 0
    if isNull(p): return 1
    return isList(schemeCdr(p))


def isPair(L):
    return type(L) in (__ListType, __TupleType)


def isDottedPair(L):
    if not isPair(L): return 0
    elif isPair(schemeCdr(L)): return 0
    else: return 1


def schemePairQuestion(L):
    return schemeBooleanize(isPair(L))


def schemeLength(p):
    if isNull(p): return 0
    return 1 + schemeLength(schemeCdr(p))


def schemeCount(x, p):
    if isNull(p): return 0
    if x == schemeCar(p):
        return 1 + schemeCount(x, schemeCdr(p))
    else:
        return schemeCount(x, schemeCdr(p))

    

def appendTwo(front, back):
    if isNull(front): return back
    return schemeCons(schemeCar(front),
                      appendTwo(schemeCdr(front), back))
    
def schemeAppend(*lists):
    appended_list = lists[-1]
    for i in range(len(lists)-2, -1, -1):
        next_list = lists[i]
        appended_list = appendTwo(next_list, appended_list)
    return appended_list



def pairsToPythonList(pair):
    """Does a shallow conversion of a pair chain to a Python list."""
    elements = []
    while not isNull(pair):
        elements.append(schemeCar(pair))
        pair = schemeCdr(pair)
    return elements


def schemeList(*elements):
    """Does a shallow conversion of a Python list to a pair chain."""
    result = list(elements)
    result.reverse()
    return reduce(lambda x, y: schemeCons(y, x), [[]] + result)


class PairTests(unittest.TestCase):
    pass


if __name__ == '__main__':
    unittest.main()
